# Source: https://github.com/SafeAILab/EAGLE/blob/main/eagle/model/ea_model.py
# Modifications: rename the entire class
import copy
import json
import time
import os

import torch
import torch.nn as nn
from transformers import PreTrainedModel, PretrainedConfig, AutoConfig
from transformers import AutoTokenizer
from huggingface_hub import hf_hub_download

from camel.models.llama.modeling_llama import LlamaForCausalLM as KVLlamaForCausalLM
from camel.utils.logits_precessor import *
from camel.utils.tree import *
from camel.utils.cache import initialize_past_key_values
from camel.utils.choices import mc_sim_7b_63
from camel.camel_modifier import CamelModifier
from camel.models.llama.configuration_llama import CamelConfig


class CamelModel(nn.Module):

    def __init__(
        self,
        base_model,
        base_model_name_or_path,
        modifier_path,
    ):

        super().__init__()
        self.base_model = base_model
        self.config = base_model.config
        self.hidden_size = base_model.lm_head.weight.shape[-1]
        self.vocab_size = base_model.lm_head.weight.shape[0]
        self.base_model_name_or_path = base_model_name_or_path
        self.tokenizer = AutoTokenizer.from_pretrained(self.base_model_name_or_path)
        config = CamelConfig.from_pretrained(modifier_path)
        with open(modifier_path, "r") as f:
            con = json.loads(f.read())
        try:
            bias = con["bias"]
        except:
            bias = True
        self.modifier = CamelModifier(config, bias=bias)

        low_memory = False

        device = base_model.model.layers[-1].self_attn.q_proj.weight.device
        if device != base_model.lm_head.weight.device:
            self.modifier.diff_device = True
            if not low_memory:
                self.modifier.headweight = base_model.lm_head.weight.clone().to(device)
            else:
                self.modifier.layer_device = device

        else:
            self.modifier.diff_device = False
        self.modifier.to(self.base_model.dtype).to(device)
        self.modifier.init_tree()

    def get_tokenizer(self):
        """Get the tokenizer of the base model.

        Returns:
            Tokenizer: The tokenizer of the base model.
        """
        return self.tokenizer

    @classmethod
    def from_pretrained(
        cls,
        type="LlamaForCausalLM",
        base_model_path=None,
        modifier_path=None,
        config_path=None,
        **kwargs,
    ):
        type = AutoConfig.from_pretrained(base_model_path).architectures[0]
        if type == "LlamaForCausalLM":
            base_model = KVLlamaForCausalLM.from_pretrained(base_model_path, **kwargs)
        else:
            raise ValueError(f"Only support base model of Llama architecture")

        if config_path is None:
            config_path = os.path.join(modifier_path, "config.json")
            if not os.path.exists(config_path):
                config_path = hf_hub_download(modifier_path, "config.json")
        model = cls(base_model, base_model_path, config_path)
        load_model_path = os.path.join(modifier_path, "pytorch_model.bin")
        if not os.path.exists(load_model_path):
            load_model_path = hf_hub_download(modifier_path, "pytorch_model.bin")
        modifier_state_dict = torch.load(
            load_model_path, map_location=base_model.device
        )
        model.modifier.load_state_dict(modifier_state_dict, strict=True)

        return model

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        labels=None,
        past_key_values=None,
        output_orig=False,
        position_ids=None,
        init=True,
        logits_processor=None,
    ):

        with torch.inference_mode():
            # Pass input through the base model
            outputs = self.base_model.model(
                input_ids=input_ids,
                attention_mask=attention_mask,
                past_key_values=past_key_values,
                position_ids=position_ids,
            )
            if output_orig:
                orig = self.base_model.lm_head(outputs[0])
            hidden_states = outputs[0].clone()
        if init:
            if logits_processor is not None:
                logits = orig[:, -1]
                logits = logits_processor(None, logits)
                probabilities = torch.nn.functional.softmax(logits, dim=1)
                token = torch.multinomial(probabilities, 1)
            else:
                token = torch.argmax(orig[:, -1])
                token = token[None, None]
            input_ids = torch.cat((input_ids, token.to(input_ids.device)), dim=1)
            # Clone the output hidden states

            camel_logits = self.modifier.topK_generate(
                hidden_states, input_ids, self.base_model.lm_head, logits_processor
            )
            if output_orig:
                return camel_logits, outputs, orig, hidden_states, token
            return camel_logits, hidden_states, token
        else:
            if output_orig:
                return outputs, orig, hidden_states

    @torch.no_grad()
    def generate(
        self,
        input_ids,
        temperature=0.0,
        top_p=0.0,
        top_k=0.0,
        max_new_tokens=512,
        max_length=2048,
        tree_choices=mc_sim_7b_63,
    ):
        if temperature > 1e-5:
            logits_processor = prepare_logits_processor(
                temperature=temperature, top_p=top_p, top_k=top_k
            )
        else:
            logits_processor = None
        # assert input_ids.shape[0] == 1, "Only support batch size 1 for now!!"
        # Avoid modifying the input_ids in-place
        input_ids = input_ids.clone()
        self.modifier.reset_kv()

        if hasattr(self, "tree_choices") and self.tree_choices == tree_choices:
            tree_buffers = self.tree_buffers
        else:
            tree_buffers = generate_tree_buffers_base(
                tree_choices,
                device=self.base_model.model.layers[-1].self_attn.q_proj.weight.device,
            )
            tree_buffers["retrieve_indices_head"] = tree_buffers["retrieve_indices"].to(
                self.base_model.lm_head.weight.device
            )
        self.tree_buffers = tree_buffers
        self.tree_choices = tree_choices

        # Initialize the past key and value states
        if hasattr(self, "past_key_values"):
            past_key_values = self.past_key_values
            past_key_values_data = self.past_key_values_data
            current_length_data = self.current_length_data
            # Reset the past key and value states
            current_length_data.zero_()
        else:
            (
                past_key_values,
                past_key_values_data,
                current_length_data,
            ) = initialize_past_key_values(self.base_model)
            self.past_key_values = past_key_values
            self.past_key_values_data = past_key_values_data
            self.current_length_data = current_length_data

        input_len = input_ids.shape[1]
        reset_tree_mode(self)
        tree_logits, logits, hidden_state, sample_token = initialize_tree(
            input_ids,
            self,
            tree_buffers["tree_attn_mask"],
            past_key_values,
            logits_processor,
        )
        new_token = 0

        for idx in range(max_length):
            candidates, cart_candidates_prob, tree_candidates = generate_candidates(
                tree_logits,
                tree_buffers["tree_indices"],
                tree_buffers["retrieve_indices"],
                sample_token,
                logits_processor,
            )
            logits, hidden_state_new, outputs = tree_decoding(
                self,
                tree_candidates,
                past_key_values,
                tree_buffers["tree_position_ids"],
                input_ids,
                tree_buffers["retrieve_indices_head"],
            )
            best_candidate, accept_length, sample_p = evaluate_posterior(
                logits,
                candidates,
                logits_processor,
                cart_candidates_prob,
                tree_logits[2],
                tree_buffers["p_indices"],
                tree_candidates,
                tree_buffers["b_indices"],
            )
            input_ids, tree_logits, new_token, hidden_state, sample_token = (
                update_inference_inputs(
                    input_ids,
                    candidates,
                    best_candidate,
                    accept_length,
                    tree_buffers["retrieve_indices"],
                    logits_processor,
                    logits,
                    tree_logits,
                    new_token,
                    past_key_values_data,
                    current_length_data,
                    self,
                    hidden_state,
                    hidden_state_new,
                    sample_p,
                )
            )

            if self.tokenizer.eos_token_id in input_ids[0, input_len:].tolist():
                return input_ids
            if new_token > max_new_tokens:
                return input_ids
            if input_ids.shape[1] > max_length:
                return input_ids
